"""Main Calendar class."""
import datetime
import logging
import os
from threading import RLock
from typing import List, Optional, Union

import yaml
from dateutil.parser import parse as dateutil_parse

logger = logging.getLogger("business")

day_interval = datetime.timedelta(days=1)
INPUT_TYPES = Union[str, datetime.date]


class Mutex:
    """Helper class for thread-safe locking."""

    def __init__(self, obj):
        """Initialise reentrant lock."""
        super().__init__()
        self.__obj = obj
        self.lock = RLock()

    def __enter__(self):
        """Acquire lock on entering."""
        self.lock.acquire()
        return self.__obj

    def __exit__(self, *args, **kwargs):
        """Release lock on exit."""
        self.lock.release()


class Calendar:
    """Calendar class."""

    _cache = Mutex(dict())

    additional_load_paths: List[str] = []
    bundled_calendars_path: List[str] = [os.path.join(os.path.dirname(__file__), "data")]

    DAY_NAMES = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
    default_working_days = ["mon", "tue", "wed", "thu", "fri"]

    def __init__(
        self,
        holidays: Optional[List[INPUT_TYPES]] = None,
        working_days: Optional[List[str]] = None,
        extra_working_dates: Optional[List[INPUT_TYPES]] = None,
    ) -> None:
        """Initialise Calendar instance."""
        self.holidays = self.parse_dates(holidays or [])
        self.working_days = [w[:3].lower() for w in working_days or self.default_working_days]
        self.extra_working_dates = self.parse_dates(extra_working_dates or [])

        # validations
        for w in self.working_days:
            if w not in self.DAY_NAMES:
                raise ValueError(f"Invalid working day name: {w}")

        for d in self.holidays:
            if d in self.extra_working_dates:
                raise ValueError(f"Holidays cannot be extra working dates: {d}")

        for d in self.extra_working_dates:
            if d.strftime("%a").lower() in self.working_days:
                raise ValueError(f"Extra working dates cannot be on working days: {d}")

    @classmethod
    def load(cls, calendar_str: str):
        """Load a scheme calendar YAML file.

        >>> %timeit -n 100 Calendar.load('bacs')
            23.9 ms ± 228 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
        """
        calendar_directories = cls.additional_load_paths + cls.bundled_calendars_path
        directory_find = [
            dir
            for dir in calendar_directories
            if os.path.exists(os.path.join(dir, f"{calendar_str}.yml"))
        ]

        if len(directory_find) >= 1:
            directory = directory_find[0]
        else:
            raise ValueError(f"No such calendar '{calendar_str}'")

        calendar_filepath = os.path.join(directory, f"{calendar_str}.yml")
        logger.debug(f"Extracting data from {calendar_filepath} yaml file")
        with open(calendar_filepath, "r") as fh:
            calendar_yaml = yaml.safe_load(fh)

        valid_keys = ["holidays", "working_days", "extra_working_dates"]
        for yaml_key in calendar_yaml.keys():
            if yaml_key not in valid_keys:
                raise ValueError(
                    f"Invalid key {yaml_key} found. Only valid keys are: {', '.join(valid_keys)}"
                )

        return cls(
            holidays=calendar_yaml.get("holidays", []),
            working_days=calendar_yaml["working_days"],
            extra_working_dates=calendar_yaml.get("extra_working_dates", []),
        )

    @classmethod
    def load_cache(cls, calendar_str: str):
        """Load a scheme calendar YAML file with cache.

        >>> %timeit Calendar.load_cache('bacs')
            969 ns ± 10.8 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
        """
        with cls._cache as cache:
            if not (calendar_str in cache):
                cache[calendar_str] = cls.load(calendar_str)
            return cache[calendar_str]

    @staticmethod
    def parse_date(input_date_raw: INPUT_TYPES) -> datetime.date:
        """Parse a raw input date.

        >>> d1 = datetime.date(2020, 1, 1)
            %timeit Calendar.parse_date(d1)
            649 ns ± 6.47 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
        >>> d2 = datetime.datetime(2020, 1, 1)
            %timeit Calendar.parse_date(d2)
            582 ns ± 17.6 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
        >>> %timeit Calendar.parse_date("2020-01-01")
            57.7 µs ± 1.47 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        """
        if isinstance(input_date_raw, datetime.datetime):
            # datetime.datetime is also an instance of datetime.date
            # this also works with pandas.Timestamp
            # so, check datetime type first!
            return input_date_raw.date()
        elif isinstance(input_date_raw, datetime.date):
            return input_date_raw
        elif isinstance(input_date_raw, str):
            return dateutil_parse(input_date_raw).date()
        else:
            raise TypeError(
                f"Unexpected input type {type(input_date_raw)} (supported: str or datetime.date)"
            )

    @staticmethod
    def parse_dates(dates: List[INPUT_TYPES]) -> List[datetime.date]:
        """Parse a list of raw input dates."""
        return [Calendar.parse_date(d) for d in dates]

    def is_holiday(self, input_date: INPUT_TYPES) -> bool:
        """Return true if the date given is a holiday."""
        input_date = self.parse_date(input_date)
        return input_date in self.holidays

    def is_working_day(self, input_date: INPUT_TYPES) -> bool:
        """Return true if the date given is a working day (typically that means a non-weekend day)."""
        input_date = self.parse_date(input_date)
        return input_date.strftime("%a").lower() in self.working_days

    def is_business_day(self, input_date: INPUT_TYPES) -> bool:
        """Return true if the date given is a working day (typically that means a non-weekend day) and not a holiday."""
        input_date = self.parse_date(input_date)
        if self.is_holiday(input_date):
            return False
        elif input_date in self.extra_working_dates:
            return True
        else:
            return self.is_working_day(input_date)

    def business_days_between(self, from_date: INPUT_TYPES, to_date: INPUT_TYPES) -> int:
        """Count the number of business days between two dates.

        This method counts from start of from_date to start of to_date. So,
        business_days_between(mon, weds) = 2 (assuming no holidays)

        To optimise this method we split the range into full weeks and a remaining period.
        We then calculate business days in the full weeks period by multiplying number of weeks by
        number of working days in a week and removing holidays one by one.

        For the remaining period, we just loop through each day and check whether it is a business day.

        >>> calendar = Calendar.load('bacs')
        >>> %timeit calendar.business_days_between(datetime.date(2020, 1, 1), datetime.date(2020, 1, 7))
            89.5 µs ± 4.3 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        >>> %timeit calendar.business_days_between(datetime.date(2020, 1, 1), datetime.date(2020, 1, 31))
            72.3 µs ± 4.84 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        >>> %timeit calendar.business_days_between(datetime.date(2020, 1, 1), datetime.date(2020, 12, 31))
            93.6 µs ± 359 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        """
        from_date = self.parse_date(from_date)
        to_date = self.parse_date(to_date)
        logger.debug(f"Calculating business days between {from_date} and {to_date}")

        # Calculate number of full weeks and remaining days
        days_between_from_to = (to_date - from_date).days
        num_full_weeks, remaining_days = divmod(days_between_from_to, 7)
        remaining_to_date = to_date - (day_interval * remaining_days)
        # First estimate for full week range based on # biz days in a week
        num_biz_days = num_full_weeks * len(self.working_days)

        # Find and remove holidays in full weeks range
        num_holidays = sum(
            1 if self.is_working_day(i) else 0
            for i in self.holidays
            if from_date <= i < remaining_to_date
        )

        # Add extra working dates in full weeks range
        num_extra_working_dates = sum(
            1 for i in self.extra_working_dates if from_date <= i < remaining_to_date
        )

        remaining_range = range((to_date - remaining_to_date).days)
        remaining_days_range = (remaining_to_date + (day_interval * i) for i in remaining_range)
        # Loop through each day in remaining_range and count if a business day
        remaining_business_days = sum(
            1 for date in remaining_days_range if self.is_business_day(date)
        )
        return num_biz_days - num_holidays + num_extra_working_dates + remaining_business_days

    def roll_forward(self, input_date: INPUT_TYPES) -> datetime.date:
        """
        Roll forward to the next business day.

        If the date given is a business day, that day will be returned.
        If the day given is a holiday ornon-working day, the next non-holiday working day will be returned.
        """
        input_date = self.parse_date(input_date)
        while not (self.is_business_day(input_date)):
            input_date += day_interval
        return input_date

    def roll_backward(self, input_date: INPUT_TYPES) -> datetime.date:
        """
        Roll backward to the previous business day.

        If the date given is a business day, that day will be returned.
        If the day given is a holiday or non-working day, the previous non-holiday working day will be returned.
        """
        input_date = self.parse_date(input_date)
        while not (self.is_business_day(input_date)):
            input_date -= day_interval
        return input_date

    def next_business_day(self, input_date: INPUT_TYPES) -> datetime.date:
        """Roll forward to the next business day regardless of whether the given date is a business day or not."""
        input_date = self.parse_date(input_date)
        input_date += day_interval
        while not (self.is_business_day(input_date)):
            input_date += day_interval
        return input_date

    def previous_business_day(self, input_date: INPUT_TYPES) -> datetime.date:
        """Roll backward to the previous business day regardless of whether the given date is a business day or not."""
        input_date = self.parse_date(input_date)
        input_date -= day_interval
        while not (self.is_business_day(input_date)):
            input_date -= day_interval
        return input_date

    def add_business_days(self, input_date: INPUT_TYPES, delta: int) -> datetime.date:
        """Add or subtract a number of business days to a date.

        If a non-business day is given, counting will start from the next business day. So:
            monday + 1 = tuesday
            friday + 1 = monday
            sunday + 1 = tuesday
            friday - 1 = thursday
            monday - 1 = friday
            sunday - 1 = thursday

        >>> calendar = Calendar.load('bacs')
        >>> %timeit calendar.add_business_days(datetime.date(2020, 1, 1), 1)
            22.6 µs ± 1.14 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        >>> %timeit calendar.add_business_days(datetime.date(2020, 1, 1), 100)
            1.05 ms ± 7.4 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
        """
        input_date = self.parse_date(input_date)
        logger.debug(f"Adding {delta} business days to {input_date}")
        if delta == 0:
            return input_date
        elif delta < 0:
            input_date = self.roll_backward(input_date)
        else:
            input_date = self.roll_forward(input_date)

        for i in range(abs(delta)):
            if delta < 0:
                input_date = self.previous_business_day(input_date)
            else:
                input_date = self.next_business_day(input_date)
        return input_date

    def get_business_day_of_month(self, input_date: INPUT_TYPES) -> int:
        """Get the business day of the month for a given input date.

        >>> calendar = Calendar.load('bacs')
        >>> %timeit calendar.get_business_day_of_month(datetime.date(2020, 1, 1))
            48.5 µs ± 621 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        >>> %timeit calendar.get_business_day_of_month(datetime.date(2020, 1, 31))
            77.2 µs ± 590 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        """
        input_date = self.parse_date(input_date)
        return self.business_days_between(input_date.replace(day=1), input_date + day_interval)



"""Init."""


def get_version():
    """Get package version."""
    try:
        from importlib.metadata import version
    except ImportError:
        from importlib_metadata import version  # type: ignore
    return version("business")


# noinspection PyBroadException
try:
    __version__ = get_version()
except Exception:
    # development mode: set dev version if package is not installed
    __version__ = "dev"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['business']

package_data = \
{'': ['*'], 'business': ['data/*']}

install_requires = \
['python-dateutil>=2.8.1,<3.0.0', 'pyyaml>=5.3.1,<6.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.6.0,<2.0.0']}

setup_kwargs = {
    'name': 'business-python',
    'version': '1.0.2',
    'description': 'Date calculations based on business calendars.',
    'long_description': '# Business (Python)\n\n[![circleci-badge](https://circleci.com/gh/gocardless/business-python.svg?style=shield)](https://app.circleci.com/pipelines/github/gocardless/business-python) [![pypi-badge](https://badge.fury.io/py/business-python.svg)](https://badge.fury.io/py/business-python)\n\nDate calculations based on business calendars. (Python 3.6+)\n\nPython implementation of https://github.com/gocardless/business\n\n## Documentation\n\nTo get business, simply:\n\n```bash\n$ pip install business-python\n```\n\n### Getting started\n\nGet started with business by creating an instance of the calendar class, passing in a hash that specifies which days of the week are considered working days, and which days are holidays.\n\n```python\nfrom business.calendar import Calendar\n\ncalendar = Calendar(\n  working_days=["monday", "tuesday", "wednesday", "thursday", "friday"],\n  # array items are either parseable date strings, or real datetime.date objects\n  holidays=["January 1st, 2020", "April 10th, 2020"],\n  extra_working_dates=[],\n)\n```\n\n`extra_working_dates` key makes the calendar to consider a weekend day as a working day.\n\nA few calendar configs are bundled with the package (see [business/data](<(business/data)>) for details). Load them by calling the `load` class method on `Calendar`.\n\n```python\ncalendar = Calendar.load("weekdays")\n```\n\nIf `working_days` is missing, then common default is used (mon-fri).\nIf `holidays` is missing, "no holidays" assumed.\nIf `extra_working_dates` is missing, then no changes in `working_days` will happen.\n\nElements of `holidays` and `extra_working_dates` may be either strings that `Calendar.parse_date()` can understand, or YYYY-MM-DD (which is considered as a Date by Python YAML itself).\n\n```yaml\nholidays:\n  - 2017-01-08 # Same as January 8th, 2017\n```\n\nThe `load_cache` method allows a thread safe way to avoid reloading the same calendar multiple times, and provides a performant way to dynamically load calendars for different requests.\n\n```python\ncalendar = Calendar.load_cache("weekdays")\n```\n\n### Input data types\n\nThe `parse_date` method is used to process the input date(s) in each method and return a `datetime.date` object.\n\n```python\nCalendar.parse_date("2019-01-01")\n# => datetime.date(2019, 1, 1)\n```\n\nSupported data types are:\n\n- `datetime.date`\n- `datetime.datetime`\n- `pandas.Timestamp` (treated as `datetime.datetime`)\n- date string parseable by [`dateutil.parser.parse`](https://dateutil.readthedocs.io/en/stable/parser.html#dateutil.parser.parse)\n\n`numpy.datetime64` is not supported, but can be converted to `datetime.date`:\n\n```python\nnumpy.datetime64(\'2014-06-01T23:00:05.453000000\').astype(\'M8[D]\').astype(\'O\')\n# =>  datetime.date(2014, 6, 1)\n```\n\n### Checking for business days\n\nTo check whether a given date is a business day (falls on one of the specified working days or extra working dates, and is not a holiday), use the `is_business_day` method on `Calendar`.\n\n```python\ncalendar.is_business_day("Monday, 8 June 2020")\n# => true\ncalendar.is_business_day("Sunday, 7 June 2020")\n# => false\n```\n\n### Custom calendars\n\nTo use a calendar you\'ve written yourself, you need to add the directory it\'s stored in as an additional calendar load path:\n\n```python\nCalendar.additional_load_paths = [\'path/to/your/calendar/directory\']\n```\n\nYou can then load the calendar as normal.\n\n### Business day arithmetic\n\n> For our purposes, date-based calculations are sufficient. Supporting time-based calculations as well makes the code significantly more complex. We chose to avoid this extra complexity by sticking solely to date-based mathematics.\n\nThe `add_business_days` method is used to perform business day arithmetic on dates.\n\n```python\ninput_date = Calendar.parse_date("Thursday, 12 June 2014")\ncalendar.add_business_days(input_date, 4).strftime("%A, %d %B %Y")\n# => "Wednesday, 18 June 2014"\ncalendar.add_business_days(input_date, -4).strftime("%A, %d %B %Y")\n# => "Friday, 06 June 2014"\n```\n\nThe `roll_forward` and `roll_backward` methods snap a date to a nearby business day. If provided with a business day, they will return that date. Otherwise, they will advance (forward for `roll_forward` and backward for `roll_backward`) until a business day is found.\n\n```python\ninput_date = Calendar.parse_date("Saturday, 14 June 2014")\ncalendar.roll_forward(input_date).strftime("%A, %d %B %Y")\n# => "Monday, 16 June 2014"\ncalendar.roll_backward(input_date).strftime("%A, %d %B %Y")\n# => "Friday, 13 June 2014"\n```\n\nIn contrast, the `next_business_day` and `previous_business_day` methods will always move to a next or previous date until a business day is found, regardless if the input provided is a business day.\n\n```python\ninput_date = Calendar.parse_date("Monday, 9 June 2014")\ncalendar.roll_forward(input_date).strftime("%A, %d %B %Y")\n# => "Monday, 09 June 2014"\ncalendar.next_business_day(input_date).strftime("%A, %d %B %Y")\n# => "Tuesday, 10 June 2014"\ncalendar.previous_business_day(input_date).strftime("%A, %d %B %Y")\n# => "Friday, 06 June 2014"\n```\n\nTo count the number of business days between two dates, pass the dates to `business_days_between`. This method counts from start of the first date to start of the second date. So, assuming no holidays, there would be two business days between a Monday and a Wednesday.\n\n```python\nfrom datetime import timedelta\n\ninput_date = Calendar.parse_date("Saturday, 14 June 2014")\ncalendar.business_days_between(input_date, input_date + timedelta(days=7))\n# => 5\n```\n\nThe `get_business_day_of_month` method return the running total of business days for a given date in that month. This method counts the number of business days from the start of the first day of the month to the given input date.\n\n```python\ninput_date = Calendar.parse_date("Thursday, 12 June 2014")\ncalendar.get_business_day_of_month(input_date)\n# => 9\n```\n\n### Included Calendars\n\nWe include some calendar data with this package but give no guarantees of its accuracy. The calendars that we include are:\n\n- ACH (United States)\n- Bacs\n- Bankgirot\n- BECS (Australia)\n- BECSNZ (New Zealand)\n- PAD (Canada)\n- Betalingsservice\n- Target (SEPA)\n- TargetFrance (SEPA + French bank holidays)\n\n## License & Contributing\n\n- This is available as open source under the terms of the [MIT License](http://opensource.org/licenses/MIT).\n- Bug reports and pull requests are welcome on GitHub at https://github.com/gocardless/business-python.\n\nGoCardless ♥ open source. If you do too, come [join us](https://gocardless.com/about/jobs).\n',
    'author': 'GoCardless',
    'author_email': 'engineering@gocardless.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gocardless/business-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
